#include "C:\Program Files\PICC\Examples\fm controler\controler.h"
#include <bh1415f.c>
#include <LCDSWEEDEN.C>
#include <LMX2322.C>
#include <internal_eeprom.c>

enum butn {null,up,down,cancel,enter};

enum menu {status1,status2,status3,settings};
enum settings_menu {freq,power,stereo,mute,parentmenu};

//lookup table for different power levels when useing an agc thats input is us and a rectified voltage power meter from the antenna with a diode drop through the recterfier of .522volts ((5/24*n+.522)/3.647)^2 and 0 when 0. might do some tests to see if theretical values match well to real life.
byte  power_lookup[16]={0,4,7,10,14,18,24,29,36,43,51,60,69,78,89,100};

//lcd char patterns (if not enough ram put in eeprom to save memeroy, or use const and change patten routine to place in rom) should really change as it uses way to much memory
byte  pattern_lr[8]={132,132,132,135,152,148,152,148}; //lr symbol
byte  pattern_l[8]={0,31,31,0,0,0,0,0};  // - symbol
byte  pattern_r[8]={0,0,0,0,0,31,31,0};  // _ symbol
byte  pattern_both[8]={0,31,31,0,0,31,31,0};  // = symbol
byte  pattern_key[8]={143,136,142,136,140,146,146,140};  // key symbol
byte  pattern_speaker[8]={129,131,157,145,145,157,131,129};  // speaker symbol
byte  pattern_mirrorspeaker[8]={144,152,151,145,145,151,152,144};  // mirrorspeaker symbol

#define LRSYMBOL 0
#define LSYMBOL 1
#define RSYMBOL 2
#define BOTHSYMBOL 3
#define KEYSYMBOL 4
#define SPEAKERSYMBOL 5
#define MIRRORSPEAKERSYMBOL 6

#define LEFTCHANNELPIN 1
#define RIGHTCHANNELPIN 0
#define PLLVOLTAGEPIN 3
#define DIVIDER 27  //(about 410/15) means volume deflection is max at about 2volts

#define DEFAULTFASTBUTONSPEED 20 // 20 -> 20*20=400ms
#define DEFAULTSLOWBUTONSPEED 48 // 48 -> 48*20=960ms

#rom 0x2100={0xFD,15,36,88,1,0} //defualt settings for freq=88.1mhz mute=off stereo=on power=100%

menu menu_ptr;
settings_menu settings_menu_ptr;
BYTE time1out;
BYTE buttontimeout;
BYTE fastbuttonspeed;
BYTE consecutivebuttonholds;
BYTE adpinptr;
butn butn_pressed;
boolean stereoon,muteon,wanted_tx_state,actual_tx_state,lockedpll;
int32 freqency;
int leftval,rightval;
int16 pllval;
int16 lastpllval; //used for simply instability test
suppress_state tx_suppress_reasion;
int txdelay_timeout;
int txpowerlevel;

//declerations
void tx_suppress(suppress_state reasion);
void tx_normal();
butn wait_for_butnhit(BYTE delay);

//interupt rutines

#INT_TIMER0
timer0(){
if(buttontimeout>0) buttontimeout--;
}

#INT_TIMER1
timer1(){
int16 tmp;
if(time1out>0) time1out--;
if(txdelay_timeout>0)txdelay_timeout--;
if(lastpllval>pllval) tmp=lastpllval-pllval; else tmp=pllval-lastpllval;
if(tmp>PLLLOCKWIDTH) tx_suppress(VOLTAGENOTSTABLE);
lastpllval=pllval;
}

#INT_TIMER2
timer2(){
read_adc(ADC_START_ONLY);
disable_interrupts(INT_TIMER2);
}

#INT_AD
adisr(){
int16 value;
value=read_adc(ADC_READ_ONLY);
if(adpinptr==LEFTCHANNELPIN)leftval=(value/DIVIDER);
if(adpinptr==RIGHTCHANNELPIN)rightval=(value/DIVIDER);

//keep an eye on the pll voltage that its not too high or too low.
if(adpinptr==PLLVOLTAGEPIN)
{
pllval=value;
if(lockedpll){
if(value<(PLLLOWERLIMIT-PLLANTIBOUNCE)) tx_suppress(FRQTOOLOW);   //if its out of range say its not locked
if(value>(PLLUPPERLIMIT+PLLANTIBOUNCE)) tx_suppress(FRQTOOHIGH);  //if its out of range say its not locked
}
else {
//if we are not locked then do
if(value<PLLLOWERLIMIT) tx_suppress(FRQTOOLOW);   //if its out of range say its not locked
if(value>PLLUPPERLIMIT) tx_suppress(FRQTOOHIGH);  //if its out of range say its not locked
if(txdelay_timeout==0) tx_normal(); //if we are not locked and we have waited a second and the pll volt is in lock range then transmit
}
}

//cycle through a0 a1 a3
if(adpinptr==3)  adpinptr=0;    //needs setting up for all ad pins every time you add a new one
 else adpinptr++;                   //remember
if(adpinptr==2) adpinptr++; //to avoid A2

set_adc_channel(adpinptr);
set_timer2(1);
enable_interrupts(INT_TIMER2);
}

void clrgohome()
{
lcd_display_clr();
lcd_cursor_gohome();
}

void clrrow(byte row)
{
int i;
lcd_cursor_goto(row,1);
for(i=1;i<=16;i++) lcd_putc(' ');
}

void entertosavetxt()
{
clrrow(2);
lcd_cursor_goto(2,1);
printf(lcd_putc,"Enter to save");
}

void entertosettxt()
{
clrrow(2);
lcd_cursor_goto(2,1);
printf(lcd_putc,"Enter to change");
}

void savedtxt()
{
clrrow(2);
lcd_cursor_goto(2,1);
printf(lcd_putc,"Saved");
wait_for_butnhit(9);
}

void tx_normal()
{
//we may have a problem when the transmitter powers up as it could cause a spike in the pll voltage causing the transmitter to go off, after a sec it goes go the off and so on. the following lines may need to be modifyed to solve this protenshal problem.
output_bit( TXONOFFPIN, wanted_tx_state);
actual_tx_state=wanted_tx_state;
lockedpll=true;
}

void tx_suppress(suppress_state reasion)
{
output_bit( TXONOFFPIN, !TXACTIVE);
actual_tx_state=!TXACTIVE;
lockedpll=false;
tx_suppress_reasion=reasion;
txdelay_timeout=TXONDELAY; // time before we consider un suppressing the signal
}

void setfreq(int32 f){
tx_suppress(CHANGINGFREQ);
delay_ms(100);
// the blancked out code below is an idea, hard to know if will have benifit in the end as its pll and vco dependent, with my set up 50khz sound a bit nicer than 100khz ref osc
//if(f==(100*(f/100)))pll_init(POSITIVE_PHASE_DETECTOR|REFERENCE_OPTIMIZATION_FREQ_5_TO_50MHZ,POWERDOWN_NONE,13000,100,f);
// else pll_init(POSITIVE_PHASE_DETECTOR|REFERENCE_OPTIMIZATION_FREQ_5_TO_50MHZ,POWERDOWN_NONE,13000,50,f);
pll_init(POSITIVE_PHASE_DETECTOR|REFERENCE_OPTIMIZATION_FREQ_5_TO_50MHZ,POWERDOWN_NONE,13000,50,f);
freqency=f;
}

void setpowerlevel(int p){
txpowerlevel=p;
if(txpowerlevel==0)wanted_tx_state=!TXACTIVE;
 else wanted_tx_state=TXACTIVE;
tx_suppress(CHANGINGPOWER);
setup_vref(VREF_LOW|txpowerlevel|VREF_A2);
}


butn getbutn()
{
butn_pressed=null;
if (!input(SW_UP))butn_pressed=up;
if (!input(SW_DOWN))butn_pressed=down;
if (!input(SW_ENTER))butn_pressed=enter;
if (!input(SW_CANCEL))butn_pressed=cancel;
if (butn_pressed!=null) {
delay_ms(100);
if(buttontimeout==0){if(consecutivebuttonholds==0)buttontimeout=DEFAULTSLOWBUTONSPEED;else buttontimeout=fastbuttonspeed;consecutivebuttonholds=1;}
 else butn_pressed=null;
}
else {buttontimeout=0;consecutivebuttonholds=0;}
return(butn_pressed);
}

butn wait_for_butnhit(BYTE delay)
{
if(delay==0) while(getbutn()==null);
else{
time1out=delay;
while((time1out>0) && (getbutn()==null) );
}
return(butn_pressed);
}

process_status_menu_butn()
{
switch (butn_pressed)
{
case cancel:
menu_ptr=status1;
break;
case up:
if(menu_ptr==settings)menu_ptr=status1;
 else menu_ptr++;
break;
case down:
if(menu_ptr==status1)menu_ptr=settings;
 else menu_ptr--;
break;
}
}

process_settings_menu_butn()
{
switch (butn_pressed)
{
case cancel:
settings_menu_ptr=parentmenu;
break;
case up:
if(settings_menu_ptr==mute)settings_menu_ptr=freq;
 else settings_menu_ptr++;
break;
case down:
if(settings_menu_ptr==freq)settings_menu_ptr=mute;
 else settings_menu_ptr--;
break;
}
}


void mnu_settings_freq()
{
int32 tmpfrq;
tmpfrq=freqency;
clrgohome();
printf(lcd_putc,"FRQ=%5.2w",freqency/10);
entertosettxt();
while(getbutn()==null);
process_settings_menu_butn();
if(butn_pressed==enter){
//start
fastbuttonspeed=0;  //fast as posible
entertosavetxt();
while(true){
switch (wait_for_butnhit(124))
{
case up:
tmpfrq+=50;
break;
case down:
tmpfrq-=50;
break;
case cancel:
fastbuttonspeed=DEFAULTFASTBUTONSPEED;
return;
case enter:
if(tmpfrq!=freqency){setfreq(tmpfrq);write_int32_eeprom(2, freqency);}
savedtxt();
fastbuttonspeed=DEFAULTFASTBUTONSPEED;
return;
case null:
fastbuttonspeed=DEFAULTFASTBUTONSPEED;
return;
}
lcd_cursor_goto(1,5);
printf(lcd_putc,"%5.2w    ",tmpfrq/10);
}
fastbuttonspeed=DEFAULTFASTBUTONSPEED;
//stop
}
}

void mnu_settings_power()
{
int tmppower;
tmppower=txpowerlevel;
clrgohome();
printf(lcd_putc,"Power=%d%%",power_lookup[tmppower]);
entertosettxt();
while(getbutn()==null);
process_settings_menu_butn();
if(butn_pressed==enter){
//start
fastbuttonspeed=10;  //a bit faster defaultfastbutonspeed/2
entertosavetxt();
while(true){
lcd_cursor_goto(1,7);
switch (wait_for_butnhit(124))
{
case up:
if(tmppower==15)tmppower=14;
tmppower+=1;
break;
case down:
if(tmppower==0)tmppower=1;
tmppower-=1;
break;
case cancel:
fastbuttonspeed=DEFAULTFASTBUTONSPEED;
return;
case enter:
fastbuttonspeed=DEFAULTFASTBUTONSPEED;
if(tmppower!=txpowerlevel){setpowerlevel(tmppower);write_eeprom (1,txpowerlevel);}
savedtxt();
return;
case null:
fastbuttonspeed=DEFAULTFASTBUTONSPEED;
return;
}
printf(lcd_putc,"%d%%   ",power_lookup[tmppower]);
}
fastbuttonspeed=DEFAULTFASTBUTONSPEED;
//stop
}
}

void mnu_settings_stereo()
{
boolean btmp;
clrgohome();
printf(lcd_putc,"Stereo=");
if(stereoon) printf(lcd_putc,"ON");
 else printf(lcd_putc,"OFF");
entertosettxt();
while(getbutn()==null);
process_settings_menu_butn();
if(butn_pressed==enter){
//start
btmp=stereoon;
entertosavetxt();
while(true){
switch (wait_for_butnhit(124))
{
case cancel:
stereo_stereo(stereoon);
return;
case enter:
if(stereoon!=btmp){stereoon=btmp;write_int1_eeprom(0, 0, stereoon);}
savedtxt();
return;
case null:
stereo_stereo(stereoon);
return;
default:
btmp=!btmp;stereo_stereo(btmp);
lcd_cursor_goto(1,8);
if(btmp) printf(lcd_putc,"ON ");
 else printf(lcd_putc,"OFF");
break;
}
}
//stop
}
}

void mnu_settings_mute()
{
boolean btmp;
clrgohome();
printf(lcd_putc,"Mute=");
if(muteon) printf(lcd_putc,"ON");
 else printf(lcd_putc,"OFF");
entertosettxt();
while(getbutn()==null);
process_settings_menu_butn();
if(butn_pressed==enter){
//start
btmp=muteon;
entertosavetxt();
while(true){
switch (wait_for_butnhit(124))
{
case cancel:
stereo_mute(muteon);
return;
case enter:
if(muteon!=btmp){muteon=btmp;write_int1_eeprom(0, 1, muteon);}
savedtxt();
return;
case null:
stereo_mute(muteon);
return;
default:
btmp=!btmp;stereo_mute(btmp);
lcd_cursor_goto(1,6);
if(btmp) printf(lcd_putc,"ON ");
 else printf(lcd_putc,"OFF");
break;
}
}
//stop
}
}

void mnu_status1()
{
int i;
boolean lastlockedpll,lastactualtxstate; //just so we dont have to print TX and keysymbol repetedly to the screen
clrgohome();
printf(lcd_putc,"%5.2wMhz ",freqency/10);
lastactualtxstate=!actual_tx_state;
lastlockedpll=!lockedpll;
lcd_cursor_goto(2,1);
if(stereoon)lcd_putc(LRSYMBOL);
 else lcd_putc(SPEAKERSYMBOL);
if(muteon)printf(lcd_putc," MUTE");
while(getbutn()==null){

if(lockedpll!=lastlockedpll) {lcd_cursor_goto(1,16);if(lockedpll)lcd_putc(KEYSYMBOL);else lcd_putc(' ');lastlockedpll=lockedpll;}
if(actual_tx_state!=lastactualtxstate) {lcd_cursor_goto(1,13);if(actual_tx_state==TXACTIVE)printf(lcd_putc,"TX");else printf(lcd_putc,"  ");lastactualtxstate=actual_tx_state;}

if(!muteon){
for(i=2;i<=16;i++){
lcd_cursor_goto(2,i);
if(stereoon){
if(((leftval)>(i-2))&&((rightval)>(i-2)))lcd_putc(BOTHSYMBOL);
if(((leftval)>(i-2))&&((rightval)<=(i-2)))lcd_putc(LSYMBOL);
if(((leftval)<=(i-2))&&((rightval)>(i-2)))lcd_putc(RSYMBOL);
if(((leftval)<=(i-2))&&((rightval)<=(i-2)))lcd_putc(' ');
}
else{
if(((leftval+rightval)/2)>(i-2))lcd_putc(255);
 else lcd_putc(' ');
}
}
}


}
process_status_menu_butn();
}

void mnu_status2()
{
if(!lockedpll){
clrgohome();
printf(lcd_putc,"VCO Problem!!\n");
if(tx_suppress_reasion==FRQTOOLOW)printf(lcd_putc,"Freq too low");
if(tx_suppress_reasion==FRQTOOHIGH)printf(lcd_putc,"Freq too high");
if(tx_suppress_reasion==VOLTAGENOTSTABLE)printf(lcd_putc,"VCO unstable");
if(wait_for_butnhit(124)==null)butn_pressed=cancel;
}
process_status_menu_butn();
/* //this is what status2 menu used to be like, the above is new
clrgohome();
printf(lcd_putc,"Stereo ");
if(stereoon)printf(lcd_putc,"ON. ");
 else printf(lcd_putc,"OFF. ");
printf(lcd_putc,"\nMute ");
if(stereoon)printf(lcd_putc,"ON.");
 else printf(lcd_putc,"OFF.");
while(getbutn()==null);
process_status_menu_butn();*/
}

void mnu_status3()
{
clrgohome();
printf(lcd_putc,"PLL Voltage");
printf(lcd_putc,"\nTX Power %d%%",power_lookup[txpowerlevel]);
do{
lcd_cursor_goto(1,13);
printf(lcd_putc,"%3.2w     ",(500*(int32)pllval)/1023);
}while(wait_for_butnhit(8)==null);
process_status_menu_butn();
}

void mnu_settings()
{
clrgohome();
printf(lcd_putc,"Settings");
entertosettxt();
if(wait_for_butnhit(124)==null)butn_pressed=cancel;
process_status_menu_butn();

if(butn_pressed==enter)
{
settings_menu_ptr=freq;
while(true){
switch (settings_menu_ptr)
{
case freq:
mnu_settings_freq();
break;
case power:
mnu_settings_power();
break;
case stereo:
mnu_settings_stereo();
break;
case mute:
mnu_settings_mute();
break;
case parentmenu:
return;
}
}
}
}



void main()
{
int i;
   port_b_pullups(TRUE);
   setup_adc_ports(ALL_ANALOG);
   setup_adc(ADC_CLOCK_INTERNAL);
   setup_spi(FALSE);
   setup_timer_0(RTCC_INTERNAL|RTCC_DIV_256);  //20.16ms till interrupt
   setup_timer_1 ( T1_INTERNAL | T1_DIV_BY_4 ); //80.66ms till interrupt;
   setup_timer_2(T2_DIV_BY_16, 0xc0, 4);  //4ms till interrupt (the smaller this is the more cpu goes to the adc rotine)
   setup_comparator(NC_NC_NC_NC);

#use fast_io(C)
#use fast_io(B)
#use fast_io(A)
SET_TRIS_B( 0xF0 );    // 0 to 3 output pins 4 to 7 input pins
SET_TRIS_A( 0xFF );    // 0 to 7 input pins
SET_TRIS_C( 0x00 );    // 0 to 7 output pins

//load settings from eeprom
stereoon=read_int1_eeprom(0,0);
muteon=read_int1_eeprom(0,1);
setpowerlevel(read_eeprom(1));  // power levels from 0 to 15 are valid 15=100% 0=0% and so on, although not linear
setfreq(read_int32_eeprom(2)); //freq in khz

enable_interrupts(INT_TIMER0);
enable_interrupts(INT_TIMER1);
enable_interrupts(INT_AD);
enable_interrupts(GLOBAL);

lcd_init(DISPLAY_ON,CURSOR_DIR_UP);
lcd_char_generator(LRSYMBOL,pattern_lr);
lcd_char_generator(LSYMBOL,pattern_l);
lcd_char_generator(RSYMBOL,pattern_r);
lcd_char_generator(BOTHSYMBOL,pattern_both);
lcd_char_generator(KEYSYMBOL,pattern_key);
lcd_char_generator(SPEAKERSYMBOL,pattern_speaker);
lcd_char_generator(MIRRORSPEAKERSYMBOL,pattern_mirrorspeaker);

stereo_stereo(stereoon);
stereo_mute(muteon);

//startup adc self perpetual cycle
adpinptr=RIGHTCHANNELPIN;
set_adc_channel(adpinptr);
read_adc(ADC_START_ONLY);

//the intro (a wast of memory)
clrgohome();
lcd_cursor_goto(1,4);
lcd_putc(MIRRORSPEAKERSYMBOL);
printf(lcd_putc," FM2006 ");
lcd_putc(SPEAKERSYMBOL);
lcd_cursor_goto(2,6);
printf(lcd_putc,"Jonti");
for(i=0;i<10;i++){
lcd_cursor_goto(1,3);
lcd_putc(LSYMBOL);
lcd_cursor_goto(1,14);
lcd_putc(RSYMBOL);
if(wait_for_butnhit(2)!=null) break;
lcd_cursor_goto(1,3);
lcd_putc(RSYMBOL);
lcd_cursor_goto(1,14);
lcd_putc(LSYMBOL);
if(wait_for_butnhit(2)!=null) break;
}

menu_ptr=status1;
fastbuttonspeed=DEFAULTFASTBUTONSPEED;  //400ms, slow is 1000ms
while(true){
switch (menu_ptr)
{
case status1:
mnu_status1();
break;
case status2:
mnu_status2();
break;
case status3:
mnu_status3();
break;
case settings:
mnu_settings();
break;
}
}


}


